/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 15/10/22.
//

#include "fggl/script/lua/engine.hpp"
#include "fggl/debug/logging.hpp"

#include "fggl/scenes/game.hpp"
#include "fggl/entity/loader/loader.hpp"

#include <cassert>

extern "C" {
	int lua_switch_scene(lua_State *L) {
		auto *scene = lua_tostring(L, -1);
		auto *statePtr = (fggl::AppState *) (lua_topointer(L, -2));
		statePtr->owner().change_state(scene);
		return 0;
	}

	int lua_create_entity(lua_State *L) {
		/*auto *prototype = lua_tostring(L, -1);
		auto *statePtr = (fggl::scenes::Game*) (lua_topointer(L, -2));

		auto *factory = statePtr->owner().service<fggl::entity::EntityFactory>();
		//factory->create( statePtr->world(), , [](){} );
		*/
		return 0;
	}

	static void open_lua_fggl(lua_State *L) {
		// switch scene
		lua_pushcfunction(L, lua_switch_scene);
		lua_setglobal(L, "switch_scene");
	}
}

namespace fggl::script::lua {

	LuaScriptProvider::LuaScriptProvider(data::Storage *storage) : m_storage(storage) {

	}

	LuaScriptEngine *LuaScriptProvider::create() {
		return new LuaScriptEngine(m_storage);
	}

	LuaScriptEngine::LuaScriptEngine(data::Storage* storage) : m_state(luaL_newstate()), m_storage(storage) {
		luaL_openlibs(m_state);
		open_lua_fggl(m_state);
	}

	LuaScriptEngine::~LuaScriptEngine() {
		release();
	}

	void LuaScriptEngine::release() {
		if ( m_state != nullptr ) {
			lua_close(m_state);
			m_state = nullptr;
		}
	}

	void LuaScriptEngine::onActivate() {
		lua_getglobal(m_state, "print");
		lua_pushstring(m_state, "LUA activate triggered");
		lua_call(m_state, 1, 0);
	}

	void LuaScriptEngine::onDeactivate() {
		lua_getglobal(m_state, "print");
		lua_pushstring(m_state, "LUA deactivate triggered");
		lua_call(m_state, 1, 0);
	}

	void LuaScriptEngine::onUpdate() {
		lua_getglobal(m_state, "print");
		lua_pushstring(m_state, "LUA update triggered");
		lua_call(m_state, 1, 0);
	}

	void LuaScriptEngine::onEvent(const std::string &name) {
		lua_getglobal(m_state, "print");
		lua_pushstring(m_state, "LUA event triggered");
		lua_call(m_state, 1, 0);
	}

	bool LuaScriptEngine::run(const char *script) {
		auto result = luaL_dostring(m_state, script);
		if ( !result ) {
			fggl::debug::warning("lua error: {}", lua_tostring(m_state, -1));
		}
		return result;
	}

	bool LuaScriptEngine::load(const char *filename) {
		assert( filename != nullptr);
		auto path = m_storage->resolvePath(data::StorageType::Data, filename);

		if ( !std::filesystem::exists(path) ) {
			fggl::debug::warning("lua error: file does not exist: {}", path.c_str());
			return false;
		}

		// load the file ( OK = 0 = false because reasons...)
		auto result = !luaL_dofile(m_state, path.c_str());
		if ( !result ) {
			fggl::debug::warning("lua error: {}", lua_tostring(m_state, -1));
			return false;
		}
		return true;
	}

	void LuaScriptEngine::setGlobal(const char *name, void *ptr) {
		lua_pushlightuserdata(m_state, ptr);
		lua_setglobal(m_state, name);
	}

}